﻿/*
 * Copyright 2021 faddenSoft
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
using System;
using System.Collections.ObjectModel;

using PluginCommon;

namespace AtariArcade {
    /// <summary>
    /// Visualization renderers for Atari Missile Command.
    ///
    /// Much of what Missile Command draws is embedded into the code (ABM icons, explosions)
    /// or with a custom data table (smartbombs, fliers).  Bitmaps are only used for alphanumeric
    /// characters, some arrows, and the cities.
    /// </summary>
    public class VisMissileCommand : MarshalByRefObject, IPlugin, IPlugin_Visualizer {
        // IPlugin
        public string Identifier {
            get { return "Missile Command Graphics"; }
        }
        private IApplication mAppRef;
        private byte[] mFileData;

        // Visualization identifiers; DO NOT change or projects that use them will break.
        private const string VIS_GEN_BITMAP_SET = "mc-bitmap-set";
        private const string VIS_GEN_FLIER = "mc-flier";
        private const string VIS_OCTAGON = "mc-octagon";

        private const string P_OFFSET = "offset";
        private const string P_COUNT = "count";
        private const string P_SHAPE_NUM = "shapeNum";
        private const string P_RADIUS = "radius";
        private const string P_SLOPE = "slope";
        private const string P_SOLID = "solid";

        private const int MAX_DIM = 4096;

        // Visualization descriptors.
        private VisDescr[] mDescriptors = new VisDescr[] {
            new VisDescr(VIS_GEN_BITMAP_SET, "8x8 Bitmap Set", VisDescr.VisType.Bitmap,
                new VisParamDescr[] {
                    new VisParamDescr("File offset (hex)",
                        P_OFFSET, typeof(int), 0, 0x00ffffff, VisParamDescr.SpecialMode.Offset, 0),
                    new VisParamDescr("Count",
                        P_COUNT, typeof(int), 1, 128, 0, 1),
                }),
            new VisDescr(VIS_GEN_FLIER, "Flier", VisDescr.VisType.Bitmap,
                new VisParamDescr[] {
                    new VisParamDescr("File offset (hex)",
                        P_OFFSET, typeof(int), 0, 0x00ffffff, VisParamDescr.SpecialMode.Offset, 0),
                    new VisParamDescr("Shape index",
                        P_SHAPE_NUM, typeof(int), 0, 1, 0, 0),
                }),
            new VisDescr(VIS_OCTAGON, "Octagon", VisDescr.VisType.Bitmap,
                new VisParamDescr[] {
                    new VisParamDescr("Radius",
                        P_RADIUS, typeof(int), 1, 127, 0, 13),
                    new VisParamDescr("Slope/8",
                        P_SLOPE, typeof(int), 1, 7, 0, 3),
                    new VisParamDescr("Solid",
                        P_SOLID, typeof(bool), false, true, 0, true),
                }),
        };


        // IPlugin
        public void Prepare(IApplication appRef, byte[] fileData, AddressTranslate addrTrans) {
            mAppRef = appRef;
            mFileData = fileData;
        }

        // IPlugin
        public void Unprepare() {
            mAppRef = null;
            mFileData = null;
        }

        // IPlugin_Visualizer
        public VisDescr[] GetVisGenDescrs() {
            if (mFileData == null) {
                return null;
            }
            return mDescriptors;
        }

        // IPlugin_Visualizer
        public IVisualization2d Generate2d(VisDescr descr,
                ReadOnlyDictionary<string, object> parms) {
            switch (descr.Ident) {
                case VIS_GEN_BITMAP_SET:
                    return GenerateBitmapSet(parms);
                case VIS_GEN_FLIER:
                    return GenerateFlier(parms);
                case VIS_OCTAGON:
                    return GenerateOctagon(parms);
                default:
                    mAppRef.ReportError("Unknown ident " + descr.Ident);
                    return null;
            }
        }

        private IVisualization2d GenerateBitmapSet(ReadOnlyDictionary<string, object> parms) {
            int offset = Util.GetFromObjDict(parms, P_OFFSET, 0);
            int count = Util.GetFromObjDict(parms, P_COUNT, 1);

            if (offset < 0 || offset >= mFileData.Length || count < 1) {
                mAppRef.ReportError("Invalid parameter");
                return null;
            }

            // Form a grid, max of 10 glyphs across.
            int hcells = (count < 10) ? count : 10;
            int vcells = (count + hcells - 1) / hcells;

            int itemByteWidth = 1;  // 1 byte across
            int itemHeight = 8;     // 8 lines high
            int cellStride = itemByteWidth * itemHeight;
            int rowStride = 1;

            // Create a bitmap with room for each cell, plus a 1-pixel transparent boundary
            // between them and around the edges.
            VisBitmap8 vb = new VisBitmap8(1 + hcells * itemByteWidth * 8 + hcells,
                                           1 + vcells * itemHeight + vcells);
            SetPalette(vb);

            int cellx = 1;
            int celly = 1;
            for (int idx = 0; idx < count; idx++) {
                RenderBitmap(offset + idx * cellStride, itemByteWidth, itemHeight, rowStride,
                    vb, cellx, celly);

                cellx += itemByteWidth * 8 + 1;
                if (cellx == vb.Width) {
                    cellx = 1;
                    celly += itemHeight + 1;
                }
            }

            return vb;
        }

        private void RenderBitmap(int offset, int byteWidth, int height, int rowStride,
                VisBitmap8 vb, int xstart, int ystart) {
            int bx = xstart;
            int by = ystart;

            // Convert bits to pixels.
            for (int row = 0; row < height; row++) {
                for (int byteCol = 0; byteCol < byteWidth; byteCol++) {
                    // Need to flip upside-down.
                    byte val = mFileData[offset + (height - row - 1) * rowStride + byteCol];
                    for (int bit = 0; bit < 8; bit++) {
                        if ((val & 0x80) != 0) {
                            vb.SetPixelIndex(bx, by, (byte)Color.Solid);
                        } else {
                            vb.SetPixelIndex(bx, by, (byte)Color.Transparent);
                        }
                        val <<= 1;
                        bx++;
                    }
                }
                bx = xstart;
                by++;
            }
        }

        // Generates data for shapes at +0011ed.  This isn't fully general, because there's
        // only one instance in the code, and I'm not sure this would be useful for anything
        // else.
        //
        // The shapes are expected to move horizontally across the screen, one pixel at a
        // time.  The leading edge is drawn in foreground colors, the trailing edge in
        // the background color, so on any given frame only a fraction of the total pixels
        // need to be drawn.
        //
        // We draw the leading/trailing pixels here, on a grey background to provide contrast
        // for the black background pixels, but it's hard to picture the full shape.  So we
        // also draw the "smeared" pixels, but with dimmed colors.
        private IVisualization2d GenerateFlier(ReadOnlyDictionary<string, object> parms) {
            int offset = Util.GetFromObjDict(parms, P_OFFSET, 0);
            int shapeNum = Util.GetFromObjDict(parms, P_SHAPE_NUM, 0);
            const int NUM_SHAPES = 2;
            const int NUM_CHUNKS = 6;

            // Crude check to avoid blowing up.
            if (offset < 0 || offset + 0xc6 >= mFileData.Length) {
                mAppRef.ReportError("Invalid parameter");
                return null;
            }
            if (shapeNum < 0 || shapeNum >= NUM_SHAPES) {
                mAppRef.ReportError("Invalid parameter");
                return null;
            }

            // Get the initial chunk index from the file data.
            int chunkIndex = mFileData[offset + shapeNum];
            if (chunkIndex >= NUM_CHUNKS) {
                mAppRef.ReportError("Invalid data: chunkIndex=$" + chunkIndex.ToString("x2"));
                return null;
            }

            // Locate the various data tables.
            offset += NUM_SHAPES;

            int dataOffsetTab = offset;
            offset += NUM_CHUNKS;
            int dataLengthTab = offset;
            offset += NUM_CHUNKS;
            int colorTab = offset;
            offset += NUM_CHUNKS;
            int horzMin = mFileData[offset + chunkIndex];
            offset += NUM_CHUNKS;
            int horzMax = mFileData[offset + chunkIndex];
            offset += NUM_CHUNKS;
            int stopFlagTab = offset;
            offset += NUM_CHUNKS;
            int horzOffsetTab = offset;
            int numPixels = mFileData[dataOffsetTab + NUM_CHUNKS - 1] +
                mFileData[dataLengthTab + NUM_CHUNKS - 1] + 1;      // should be 80
            offset += numPixels;
            int vertOffsetTab = offset;

            // We can detect the max width from the horizontal min/max table entries, but
            // we can't get the vertical limit without scanning the data.  For simplicity
            // we scan the entire table, but for correct behavior we should only be scanning
            // the chunks used in the specific shape.
            int minX = 1000;
            int minY = 1000;
            int maxX = -1000;
            int maxY = -1000;
            for (int i = 0; i < numPixels; i++) {
                int xc = (sbyte)mFileData[horzOffsetTab + i];
                if (xc < minX) {
                    minX = xc;
                }
                if (xc > maxX) {
                    maxX = xc;
                }

                int yc = (sbyte)mFileData[vertOffsetTab + i];
                if (yc < minY) {
                    minY = yc;
                }
                if (yc > maxY) {
                    maxY = yc;
                }
            }
            int width = maxX - minX;
            int height = maxY - minY;
            //mAppRef.DebugLog("width=" + width + " height=" + height +
            //    " (horzMin=" + horzMin + " horzMax=" + horzMax + ")");

            byte[,] colorGrid = new byte[height + 3, width + 3];
            VisBitmap8 vb = new VisBitmap8(width + 3, height + 3);
            SetPalette(vb);
            vb.SetAllPixelIndices((byte)Color.FlierBk);

            int centerX = width / 2 + 2;
            int centerY = height / 2 + 1;

            //
            // Draw the data from the game.  This is just the leading and trailing edges.
            //
            int contFlag;
            do {
                int dataOffset = mFileData[dataOffsetTab + chunkIndex];
                int dataLength = mFileData[dataLengthTab + chunkIndex] + 1;
                int colorIndex = mFileData[colorTab + chunkIndex];
                byte color;
                switch (colorIndex) {
                    case 0x00: color = (byte)Color.Flier0; break;
                    case 0x40: color = (byte)Color.Flier2; break;
                    case 0x80: color = (byte)Color.Flier4; break;
                    case 0xe0: color = (byte)Color.Flier6; break;
                    default: color = (byte)Color.Solid; break;    // not expected
                }
                while (dataLength-- != 0) {
                    int xc = (sbyte)mFileData[horzOffsetTab + dataOffset];
                    int yc = (sbyte)mFileData[vertOffsetTab + dataOffset];
                    vb.SetPixelIndex(centerX + xc, centerY - yc, color);
                    colorGrid[centerY - yc, centerX + xc] = color;
                    dataOffset++;
                }

                contFlag = mFileData[stopFlagTab + chunkIndex];
                chunkIndex--;
            } while (contFlag == 0);

            // We recorded the colors in a separate array because vb.GetPixelIndex() didn't exist
            // when this was first written.  We use those values to generate the "smeared" pixels
            // that would appear when the shape is animated.
            //
            // Note unset values are 0, rather than Color.FlierBk, in the array.
            for (int row = 0; row < vb.Height; row++) {
                byte latchedColor = 0;
                // Shape is drawn as if moving left to right, so scan right to left.
                for (int col = vb.Width - 1; col >= 0; col--) {
                    byte bitmapColor = colorGrid[row, col]; // vb.GetPixelIndex(col, row);
                    if (bitmapColor == 0 || bitmapColor == (byte)Color.FlierBk) {
                        // No color set here.  If we've latched a color, output the "dim" version.
                        if (latchedColor != 0) {
                            vb.SetPixelIndex(col, row, latchedColor);
                        }
                    } else if (bitmapColor == (byte)Color.Flier0) {
                        // Found a background-color pixel.  Clear the latched value.
                        latchedColor = 0;
                    } else {
                        // Found a foreground-color pixel.  Latch the "dim" version.
                        latchedColor = (byte)(bitmapColor + 1);
                    }
                }
            }

            return vb;
        }

        private IVisualization2d GenerateOctagon(ReadOnlyDictionary<string, object> parms) {
            int radius = Util.GetFromObjDict(parms, P_RADIUS, 13);
            int slope = Util.GetFromObjDict(parms, P_SLOPE, 3);
            bool solid = Util.GetFromObjDict(parms, P_SOLID, true);

            // The table at $54b7 uses $0d as the largest radius for an explosion.  The
            // game-over code at $5da2 goes out to 108 for "the end".
            if (radius < 1 || radius > 127 || slope < 1 || slope > 7) {
                mAppRef.ReportError("Invalid parameter");
                return null;
            }

            int dim = 32;
            int center = 15;
            if (radius > 15) {
                dim = 256;
                center = 127;
            }

            VisBitmap8 vb = new VisBitmap8(dim, dim);
            SetPalette(vb);

            if (solid) {
                for (int rad = 0; rad <= radius; rad++) {
                    DrawOctagon(vb, center, rad, slope);
                }
            } else {
                //DrawOctagon(vb, center, 0, slope);
                DrawOctagon(vb, center, radius, slope);
            }

            return vb;
        }

        /// <summary>
        /// Draws an octagon using the game's algorithm.  The only difference is that
        /// the slope may be something other than 3/8.
        /// </summary>
        private void DrawOctagon(VisBitmap8 vb, int center, int radius, int slope) {
            int deltaX = radius;
            int deltaY = 0;

            while (deltaX >= deltaY) {
                DrawFourPix(vb, center, deltaX, deltaY);
                int tmp = deltaX;
                deltaX = deltaY;
                deltaY = tmp;
                DrawFourPix(vb, center, deltaX, deltaY);

                deltaY = deltaX + 1;
                deltaX = radius - (deltaY * slope) / 8;
            }
        }

        private void DrawFourPix(VisBitmap8 vb, int center, int deltaX, int deltaY) {
            byte color = (byte)Color.Solid;

            vb.SetPixelIndex(center + deltaX, center + deltaY, color);
            vb.SetPixelIndex(center - deltaX, center + deltaY, color);
            vb.SetPixelIndex(center - deltaX, center - deltaY, color);
            vb.SetPixelIndex(center + deltaX, center - deltaY, color);
        }

        private enum Color : byte {
            //FramingBackground = 0,
            Transparent = 0,
            Solid = 1,

            // Colors for the bomber/satellite rendering.
            FlierBk = 2,
            Flier0 = 3,
            Flier2 = 4,
            Flier2dim = 5,
            Flier4 = 6,
            Flier4dim = 7,
            Flier6 = 8,
            Filer6dim = 9,
        }

        private void SetPalette(VisBitmap8 vb) {
            //vb.AddColor(0x40, 0, 0, 0);             // framing background
            vb.AddColor(0, 0, 0, 0);                // 0=transparent
            vb.AddColor(0xff, 0x20, 0x20, 0xff);    // 1=solid

            vb.AddColor(0xff, 0x80, 0x80, 0x80);    // 2=FlierBk (grey)
            vb.AddColor(0xff, 0x00, 0x00, 0x00);    // 3=Flier0 (background / black)
            vb.AddColor(0xff, 0xff, 0x00, 0x00);    // 4=Flier2 (red)
            vb.AddColor(0xff, 0xaf, 0x00, 0x00);    // 5=Flier2dim (dim red)
            vb.AddColor(0xff, 0xff, 0xa5, 0x00);    // 6=Flier4 (cycling / orange)
            vb.AddColor(0xff, 0xaf, 0x50, 0x00);    // 7=Flier4dim (dim orange)
            vb.AddColor(0xff, 0x00, 0x00, 0xff);    // 8=Flier6 (blue)
            vb.AddColor(0xff, 0x00, 0x00, 0xaf);    // 9=Flier6dim (dim blue)
        }
    }
}
